@@ -1,6 +1,6 @@
 # sdtfile.py
 
-# Copyright (c) 2007-2021, Christoph Gohlke
+# Copyright (c) 2007-2022, Christoph Gohlke
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -39,59 +39,85 @@ instrumentation parameters and measurement data. Currently only the
 `Becker & Hickl GmbH <http://www.becker-hickl.de/>`_ is a manufacturer of
 equipment for photon counting.
 
-:Author:
-  `Christoph Gohlke <https://www.lfd.uci.edu/~gohlke/>`_
-
-:Organization:
-  Laboratory for Fluorescence Dynamics. University of California, Irvine
-
+:Author: `Christoph Gohlke <https://www.cgohlke.com>`_
 :License: BSD 3-Clause
-
-:Version: 2021.11.18
+:Version: 2022.9.28
 
 Requirements
 ------------
-* `CPython >= 3.7 <https://www.python.org>`_
-* `Numpy 1.19 <https://www.numpy.org>`_
+
+This release has been tested with the following requirements and dependencies
+(other versions may work):
+
+- `CPython 3.8.10, 3.9.13, 3.10.7, 3.11.0rc2 <https://www.python.org>`_
+- `Numpy 1.22.4 <https://pypi.org/project/numpy/>`_
 
 Revisions
 ---------
+
+2022.9.28
+
+- Convert docstrings to Google style with Sphinx directives.
+
+2022.2.2
+
+- Add type hints.
+- Drop support for Python 3.7 and numpy < 1.19 (NEP29).
+
 2021.11.18
-    Fix reading FLIM files created by Prairie View software (#5).
+
+- Fix reading FLIM files created by Prairie View software (#5).
+
 2021.3.21
-    Add sdt2dat script.
+
+- Add sdt2dat script.
+
 2020.12.10
-    Fix shape of non-square frames.
+
+- Fix shape of non-square frames.
+
 2020.8.3
-    Fix integer overflow (#3).
-    Support os.PathLike file names.
+
+- Fix integer overflow (#3).
+- Support os.PathLike file names.
+
 2020.1.1
-    Fix reading MCS_BLOCK data.
-    Remove support for Python 2.7 and 3.5.
-    Update copyright.
+
+- Fix reading MCS_BLOCK data.
+- Remove support for Python 2.7 and 3.5.
+- Update copyright.
+
 2019.7.28
-    Fix reading compressed, multi-channel data.
+
+- Fix reading compressed, multi-channel data.
+
 2018.9.22
-    Use str, not bytes for ASCII data.
+
+- Use str, not bytes for ASCII data.
+
 2018.8.29
-    Move module into sdtfile package.
+
+- Move module into sdtfile package.
+
 2018.2.7
-    Bug fixes.
+
+- Bug fixes.
+
 2016.3.30
-    Support revision 15 files and compression.
+
+- Support revision 15 files and compression.
+
 2015.1.29
-    Read SPC DLL data files.
+
+- Read SPC DLL data files.
+
 2014.9.5
-    Fix reading multiple MEASURE_INFO records.
 
-Notes
------
-The API is not stable yet and might change between revisions.
-
-This module has been tested with a limited number of files only.
+- Fix reading multiple MEASURE_INFO records.
 
 References
 ----------
+
 1. W Becker. The bh TCSPC Handbook. Third Edition. Becker & Hickl GmbH 2008.
    pp 401.
 2. SPC_data_file_structure.h header file. Part of the Becker & Hickl
@@ -99,6 +125,7 @@ References
 
 Examples
 --------
+
 Read image and metadata from a "SPC Setup & Data File":
 
 >>> sdt = SdtFile('image.sdt')
@@ -139,20 +166,23 @@ Read image data from a "SPC FCS Data File" as numpy array:
 
 """
 
-__version__ = '2021.11.18'
+from __future__ import annotations
 
-__all__ = (
+__version__ = '2022.9.28'
+
+__all__ = [
     'SdtFile',
     'FileInfo',
     'SetupBlock',
     'BlockNo',
     'BlockType',
     'FileRevision',
-)
+]
 
 import os
 import io
 import zipfile
+from typing import BinaryIO
 
 import numpy
 
@@ -160,40 +190,42 @@ import numpy
 class SdtFile:
     """Becker & Hickl SDT file.
 
-    Attributes
-    ----------
-    header : numpy.rec.array of FILE_HEADER structure
-        General information about the location of the setup and measurement
-        data within the file.
-    info : FileInfo
-        General information in ASCII format.
-    setup : SetupBlock or None
-        Setup block containing all system parameters, display parameters, etc.
-    measure_info : numpy.rec.array of MEASURE_INFO structure
-        Measurement description blocks.
-    block_headers : list of numpy.rec.array of BLOCK_HEADER structure
-        Data block headers.
-    data : list of 2D numpy arrays
-        Photon counts at each curve point.
-    times : list of 1D numpy arrays
-        Time axes for each data set.
+    Parameters:
+        arg: File name or open file.
 
     """
 
-    def __init__(self, arg):
-        """Initialize instance from file name or open file."""
-        if hasattr(arg, 'seek'):
-            self.filename = ''
-            self._fromfile(arg)
-        else:
+    filename: str
+    """Name of file."""
+    header: numpy.recarray
+    """File header of type FILE_HEADER."""
+    info: FileInfo
+    """File info string and attributes."""
+    setup: SetupBlock | None
+    """Setup block ascii and binary data."""
+    measure_info: list[numpy.recarray]
+    """Measurement description blocks of type MEASURE_INFO."""
+    block_headers: list[numpy.recarray]
+    """Data block headers of type BLOCK_HEADER."""
+    data: list[numpy.ndarray]
+    """Photon counts at each curve point."""
+    times: list[numpy.ndarray]
+    """Time axes for each data set."""
+
+    def __init__(self, arg: str | os.PathLike | BinaryIO, /) -> None:
+        if isinstance(arg, (str, os.PathLike)):
             self.filename = os.fspath(arg)
             with open(arg, 'rb') as fh:
                 self._fromfile(fh)
+        else:
+            assert hasattr(arg, 'seek')
+            self.filename = ''
+            self._fromfile(arg)
 
-    def _fromfile(self, fh):
+    def _fromfile(self, fh: BinaryIO, /) -> None:
         """Initialize instance from open file."""
         # read file header
-        self.header = numpy.rec.fromfile(
+        self.header = numpy.rec.fromfile(  # type: ignore
             fh, dtype=FILE_HEADER, shape=1, byteorder='<'
         )[0]
         if self.header.header_valid != 0x5555:
@@ -213,8 +245,9 @@ class SdtFile:
                 'SPC Setup & Data File',
                 'SPC FCS Data File',
                 'SPC DLL Data File',
+                'SPC Setup & Data File',  # corrupted?
             ):
-                raise NotImplementedError('not supported:', self.info.id)
+                raise NotImplementedError(f'{self.info.id!r} not supported')
         except AttributeError as exc:
             raise ValueError('invalid SDT file info\n', self.info) from exc
 
@@ -235,7 +268,9 @@ class SdtFile:
         fh.seek(self.header.meas_desc_block_offset)
         for _ in range(self.header.no_of_meas_desc_blocks):
             self.measure_info.append(
-                numpy.rec.fromfile(fh, dtype=dtype, shape=1, byteorder='<')
+                numpy.rec.fromfile(  # type: ignore
+                    fh, dtype=dtype, shape=1, byteorder='<'
+                )
             )
             fh.seek(self.header.meas_desc_block_length - dtype.itemsize, 1)
 
@@ -250,7 +285,7 @@ class SdtFile:
         for _ in range(self.header.no_of_data_blocks):
             # read data block header
             fh.seek(offset)
-            bh = numpy.rec.fromfile(
+            bh = numpy.rec.fromfile(  # type: ignore
                 fh, dtype=block_header_t, shape=1, byteorder='<'
             )[0]
             self.block_headers.append(bh)
@@ -263,9 +298,9 @@ class SdtFile:
             if bt.compress:
                 bio = io.BytesIO(fh.read(bh.next_block_offs - bh.data_offs))
                 with zipfile.ZipFile(bio) as zf:
-                    data = zf.read(zf.filelist[0].filename)  # 'data_block'
+                    databytes = zf.read(zf.filelist[0].filename)  # data_block
                 del bio
-                data = numpy.frombuffer(data, dtype=dtype, count=dsize)
+                data = numpy.frombuffer(databytes, dtype=dtype, count=dsize)
             else:
                 data = numpy.fromfile(fh, dtype=dtype, count=dsize)
 
@@ -296,44 +331,55 @@ class SdtFile:
             self.times.append(t)
             offset = bh.next_block_offs
 
-    def block_measure_info(self, block):
-        """Return measure_info record for data block."""
+    def block_measure_info(self, block: int, /) -> numpy.recarray:
+        """Return measure_info record for data block.
+
+        Parameters:
+            block: Block index.
+
+        """
         return self.measure_info[self.block_headers[block].meas_desc_block_no]
 
-    def __enter__(self):
+    def __enter__(self) -> SdtFile:
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
         pass
 
-    def __str__(self):
-        """Return string containing all information about SdtFile."""
-        # measure_info = '\n  '.join(str(i) for i in self.measure_info)
-        # block_headers = '\n  '.join(
-        #     indent(i, '  ') for i in self.block_headers)
-        block_types = '\n '.join(
-            indent(BlockType(i.block_type)) for i in self.block_headers
-        )
-        shapes = '\n  '.join(str(i.shape) for i in self.data)
-        return '\n '.join(
-            (
-                self.__class__.__name__,
-                os.path.normpath(os.path.normcase(self.filename)),
-                indent(FileRevision(self.header.revision)),
-                indent(self.info.strip()),
-                # f'header: {self.header}',
-                # f'measure_info:\n  {measure_info}',
-                # f'block_headers:\n  {block_headers}',
-                block_types,
-                f'shapes:\n  {shapes}',
-            )
+    def __repr__(self) -> str:
+        filename = os.path.split(self.filename)[-1]
+        return f'<{self.__class__.__name__} {filename!r}>'
+
+    def __str__(self) -> str:
+        return indent(
+            repr(self),
+            # os.path.normpath(os.path.normcase(self.filename)),
+            FileRevision(self.header.revision),
+            indent('info:', self.info.strip()),
+            # indent('header:', self.header),
+            # indent('measure_info:', *self.measure_info),
+            # indent('block_headers:', *self.block_headers),
+            indent(
+                'blocktypes:',
+                *(BlockType(i.block_type) for i in self.block_headers),
+            ),
+            indent('shapes:', *(i.shape for i in self.data)),
         )
 
 
 class FileInfo(str):
-    """File info string and attributes."""
+    """File info string and attributes.
 
-    def __init__(self, value):
+    Parameters:
+        value:
+            File content from FILE_HEADER info_offset and info_length.
+
+    """
+
+    id: str
+    """Identification."""
+
+    def __init__(self, value: str, /) -> None:
         str.__init__(self)
         assert value.startswith('*IDENTIFICATION') and value.strip().endswith(
             '*END'
@@ -349,11 +395,22 @@ class FileInfo(str):
 
 
 class SetupBlock:
-    """Setup block ascii and binary data."""
+    """Setup block ascii and binary data.
+
+    Parameters:
+        value:
+            File content from FILE_HEADER setup_offs and setup_length.
+
+    """
 
     __slots__ = ('ascii', 'binary')
 
-    def __init__(self, value):
+    ascii: str
+    """ASCII data."""
+    binary: bytes | None
+    """Binary data."""
+
+    def __init__(self, value: bytes, /) -> None:
         assert value.startswith(b'*SETUP') and value.strip().endswith(b'*END')
         i = value.find(b'BIN_PARA_BEGIN')
         if i:
@@ -364,64 +421,87 @@ class SetupBlock:
             self.ascii = value.decode('windows-1250')
             self.binary = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.ascii
 
 
 class BlockNo:
-    """The lblock_no field of BLOCK_HEADER."""
+    """BLOCK_HEADER.lblock_no field.
+
+    Parameters:
+        value: Value of BLOCK_HEADER.lblock_no.
+
+    """
 
     __slots__ = ('data', 'module')
 
-    def __init__(self, value):
+    data: int
+    """Data."""
+    module: int
+    """Module."""
+
+    def __init__(self, value: int, /) -> None:
         self.data = (value & 0xFFFFFF00) >> 24
         self.module = value & 0x000000FF
 
-    def __iter__(self):
-        return iter((self.data, self.module))
-
-    def __str__(self):
-        return '\n '.join(
-            (
-                self.__class__.__name__,
-                f'data number: {self.data}',
-                f'module number: {self.module}',
-            )
-        )
+    def __repr__(self) -> str:
+        return f'<{self.__class__.__name__} {self.data} {self.module}>'
 
 
 class BlockType:
-    """The block_type field of BLOCK_HEADER."""
+    """BLOCK_HEADER.block_type field.
+
+    Parameters:
+        value: Value of BLOCK_HEADER.block_type.
+
+    """
 
     __slots__ = ('mode', 'contents', 'dtype', 'compress')
 
-    def __init__(self, value):
+    mode: str
+    """BLOCK_CREATION."""
+    contents: str
+    """BLOCK_CONTENT."""
+    dtype: numpy.dtype
+    """BLOCK_DTYPE."""
+    compress: bool
+    """Data is compressed."""
+
+    def __init__(self, value: int, /) -> None:
         self.mode = BLOCK_CREATION[value & 0xF]
         self.contents = BLOCK_CONTENT[value & 0xF0]
         self.dtype = BLOCK_DTYPE[value & 0xF00]
         self.compress = bool(value & 0x1000)
 
-    def __iter__(self):
-        return iter((self.mode, self.contents, self.dtype))
+    def __repr__(self) -> str:
+        return f'<{self.__class__.__name__} {self.mode} {self.contents}>'
 
-    def __str__(self):
-        return '\n '.join(
-            (
-                self.__class__.__name__,
-                f'mode: {self.mode}',
-                f'contents: {self.contents}',
-                f'dtype: {self.dtype}',
-                f'compress: {self.compress}',
-            )
+    def __str__(self) -> str:
+        return indent(
+            repr(self),
+            # f'mode: {self.mode}',
+            # f'contents: {self.contents}',
+            f'dtype: {self.dtype}',
+            f'compress: {self.compress}',
         )
 
 
 class FileRevision:
-    """The revision field of FILE_HEADER."""
+    """FILE_HEADER.revision field.
+
+    Parameters:
+        value: Value of FILE_HEADER.revision.
+
+    """
 
     __slots__ = ('revision', 'module')
 
-    def __init__(self, value):
+    revision: int
+    """Revision."""
+    module: str
+    """Module."""
+
+    def __init__(self, value: int, /) -> None:
         self.revision = value & 0b1111
         self.module = {
             0x20: 'SPC-130',
@@ -437,17 +517,13 @@ class FileRevision:
             0x2A: 'SPC-130EM',
         }.get((value & 0xFF0) >> 4, 'Unknown')
 
-    def __str__(self):
-        return '\n '.join(
-            (
-                self.__class__.__name__,
-                f'revision: {self.revision}',
-                f'module type: {self.module}',
-            )
+    def __repr__(self) -> str:
+        return (
+            f'<{self.__class__.__name__} {self.module!r} rev {self.revision}>'
         )
 
 
-FILE_HEADER = [
+FILE_HEADER: list[tuple[str, str]] = [
     ('revision', 'i2'),
     ('info_offset', 'i4'),
     ('info_length', 'i2'),
@@ -465,7 +541,7 @@ FILE_HEADER = [
     ('chksum', 'u2'),
 ]
 
-SETUP_BIN_HDR = [
+SETUP_BIN_HDR: list[tuple[str, str]] = [
     ('soft_rev', 'u4'),
     ('para_length', 'u4'),
     ('reserved1', 'u4'),
@@ -473,7 +549,7 @@ SETUP_BIN_HDR = [
 ]
 
 # Info collected when measurement finished
-MEASURE_STOP_INFO = [
+MEASURE_STOP_INFO: list[tuple[str, str]] = [
     ('status', 'u2'),
     ('flags', 'u2'),
     ('stop_time', 'f4'),
@@ -493,7 +569,7 @@ MEASURE_STOP_INFO = [
 ]
 
 # Info collected when FIFO measurement finished
-MEASURE_FCS_INFO = [
+MEASURE_FCS_INFO: list[tuple[str, str]] = [
     ('chan', 'u2'),
     ('fcs_decay_calc', 'u2'),
     ('mt_resol', 'u4'),
@@ -510,7 +586,7 @@ MEASURE_FCS_INFO = [
 ]
 
 # Extension of MeasFCSInfo for other histograms
-HIST_INFO = [
+HIST_INFO: list[tuple[str, str]] = [
     ('fida_time', 'f4'),
     ('filda_time', 'f4'),
     ('fida_points', 'i4'),
@@ -520,7 +596,7 @@ HIST_INFO = [
 ]
 
 # Measurement description blocks
-MEASURE_INFO = [
+MEASURE_INFO: list[tuple[str, str | list[tuple[str, str]]]] = [
     ('time', 'a9'),
     ('date', 'a11'),
     ('mod_ser_no', 'a16'),
@@ -592,7 +668,7 @@ MEASURE_INFO = [
     ('MeasHISTInfo', HIST_INFO),
 ]
 
-BLOCK_HEADER = [
+BLOCK_HEADER: list[tuple[str, str]] = [
     ('block_no', 'i2'),
     ('data_offs', 'i4'),
     ('next_block_offs', 'i4'),
@@ -602,7 +678,7 @@ BLOCK_HEADER = [
     ('block_length', 'u4'),
 ]
 
-BLOCK_HEADER_15 = [
+BLOCK_HEADER_15: list[tuple[str, str]] = [
     ('data_offs_ext', 'u1'),
     ('next_block_offs_ext', 'u1'),
     ('data_offs', 'u4'),
@@ -614,7 +690,7 @@ BLOCK_HEADER_15 = [
 ]
 
 # Mode of creation
-BLOCK_CREATION = {
+BLOCK_CREATION: dict[int, str] = {
     0: 'NOT_USED',
     1: 'MEAS_DATA',
     2: 'FLOW_DATA',
@@ -625,7 +701,7 @@ BLOCK_CREATION = {
     9: 'FIFO_DATA_FROM_FILE',
 }
 
-BLOCK_CONTENT = {
+BLOCK_CONTENT: dict[int, str] = {
     0x0: 'DECAY_BLOCK',
     0x10: 'PAGE_BLOCK',
     0x20: 'FCS_BLOCK',
@@ -642,15 +718,15 @@ BLOCK_CONTENT = {
 }
 
 # Data type
-BLOCK_DTYPE = {
+BLOCK_DTYPE: dict[int, numpy.dtype] = {
     0x000: numpy.dtype('<u2'),
     0x100: numpy.dtype('<u4'),
     0x200: numpy.dtype('<f8'),
 }
 
-HEADER_VALID = {0x1111: False, 0x5555: True}
+HEADER_VALID: dict[int, bool] = {0x1111: False, 0x5555: True}
 
-INFO_IDS = {
+INFO_IDS: dict[str, str] = {
     'SPC Setup Script File': 'Setup script mode: setup only',
     'SPC Setup & Data File': 'Normal mode: setup + data',
     'SPC DLL Data File': 'DLL created: no setup, only data',
@@ -662,9 +738,12 @@ INFO_IDS = {
 }
 
 
-def indent(string, pad=' '):
-    """Return string with all but first line indented."""
-    return f'\n{pad}'.join(str(string).splitlines())
+def indent(*args) -> str:
+    """Return joined string representations of objects with indented lines."""
+    text = '\n'.join(str(arg) for arg in args)
+    return '\n'.join(
+        ('  ' + line if line else line) for line in text.splitlines() if line
+    )[2:]
 
 
 if __name__ == '__main__':
